/* $Revision: 1.15 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGridPds.C
 *
 * DESCRIPTION: Fill in an XcGrid from a PDS file.  This is a JPL/MIPL
 * 	image format
 *
 * EXPORTS:  
 *
 * NOTES:        
 *
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

/***************************************************************************
 * INCLUDES
 ***************************************************************************/

//
// C headers
//
#include <cassert>
#include <cstdlib>

//
// C++ headers
//
#include <iostream>
#include <fstream>
#include <string>

//
// Local headers
//
#include "XcGrid.H"	// Already included from XcGridPds.H
#include "XcGridPds.H"
#include "PdsLabel.H"	// Handles the Pds Labels

/***************************************************************************
 * FUNCTIONS AND DEFINITIONS
 ***************************************************************************/

enum dataType {BYTE, HALF, FULL, REAL, DOUB, COMP, UNKNOWN_DATA_TYPE};
enum endian {LITTLE,BIG,UNKNOWN_ENDIAN}; // aka LOW, HIGH

static dataType getDataType (PdsLabelMap &labels) {
  //PdsLabel type = labels ["IMAGE-SAMPLE_TYPE"]; // MSB-INTEGER
  const int sbits = labels["IMAGE-SAMPLE_BITS"].getIntVal();
  if (8==sbits) return BYTE;
  if (16==sbits) return HALF;
  std::cout << "FIX: DANGER WILL ROBINSON \n"
	    << "   need to look at the sample type too...\n";
  if (128==sbits) return FULL; // 
  //if ("BYTE" == f) return BYTE;
  //if ("HALF" == f) return HALF;
  //if ("FULL" == f) return FULL;
  //if ("REAL" == f) return REAL;
  //if ("DOUB" == f) return DOUB;
  //if ("COMP" == f) return COMP;
  //std::cerr << "ERROR: getDataType does not recognize this type..."
  //	    << f << "\n";
  return (UNKNOWN_DATA_TYPE);
}

static endian getDataEndian(PdsLabelMap &labels) {
  const int sbits = labels["IMAGE-SAMPLE_BITS"].getIntVal();
  if (8==sbits) return LITTLE; // Doesn't matter

  std::string type = labels ["IMAGE-SAMPLE_TYPE"].getStringVal(); // MSB-INTEGER

  if (0==type.compare(0,3,"MSB")) {
    //std::cout << "MSB for " << type << "\n";
    return(BIG);
  }
  if (0==type.compare(0,3,"LSB")) {
    //std::cout << "LSB for " << type << "\n";
    return(LITTLE);
  }
  
  //assert(false);
  //if ("LOW"==intfmt) return (LITTLE);
  //if ("HIGH"==intfmt) return (BIG);
  //std::cerr << "Encountered unknown endian type: " << intfmt << "\n";
  std::cerr << "FIX: check endian for "<<type<<"\n";
  return (UNKNOWN_ENDIAN);
}

// FIX: move this somewhere else and make sure there is only one copy!
// Also in the Vicar code
static endian getHostEndian(void) {
  int short s=1;
  char *b = (char *)&s;
  if (b[0]) return (LITTLE);
  return (BIG);
}
// Number of bytes per sample
static int getDataValueSize(const dataType t) {
  switch(t) {
  case BYTE:
    return(1);
  case HALF:
    return(2);
  case FULL:
    return(4);
  case REAL:
    return(4);
  case DOUB:
    return(8);
  case COMP:
    return(8); // FIX: is this right??
  case UNKNOWN_DATA_TYPE:
    std::cerr << "ERROR: asking for size on UNKNOWN_DATA_TYPE.  0 is what you get.\n";
    return(0);
  default:
    std::cerr << "ERROR: how the heck did we get here?\n";
  }
  return(-1);
}
void fixEndian(short &s,const endian e, const endian he) {
  if (e == he) return;
  char *b = (char *)&s;
  char tmp;
  tmp=b[0];  b[0]=b[1]; b[1]=tmp;
}

void fixEndian(int &i,const endian e, const endian he) {
  if (e == he) return;
  char *b = (char *)&i;
  char tmp[4];
  int x;
  for (x=0;x<4;x++) tmp[x]=b[x];
  for (x=0;x<4;x++) b[x]=tmp[3-x];
}

void fixEndian(float &f,const endian e, const endian he) {
  if (e == he) return;
  char *b = (char *)&f;
  char tmp[4];
  int x;
  for (x=0;x<4;x++) tmp[x]=b[x];
  for (x=0;x<4;x++) b[x]=tmp[3-x];
}

// Number of bytes per sample
float getFloatVal(char *ptr, const dataType t, const endian e) {
  static endian he = getHostEndian();
  switch(t) {
  case BYTE:
    {
      // No endianness to worry about.  Assum
      unsigned char *v = (unsigned char *)ptr;
      return(*v);
    }
  case HALF:
    {
      short s = *(short *)ptr;
      fixEndian(s,e,he);
      return(s);
    }
  case FULL:
    {
      int i = *(int *)ptr;
      fixEndian(i,e,he);
      return(i);
    }
  case REAL:
    {
      float f = *(float *)ptr;
      fixEndian(f,e,he);
      return(f);
    }
  case DOUB:
    std::cerr << "FIX: need to implement reading DOUB format data\n";
    assert(false);
  case COMP:
    std::cerr << "FIX: need to implement reading COMP format data\n";
    // FIX: what the heck do I do with complex?
    assert(false);
  case UNKNOWN_DATA_TYPE:
    std::cerr << "ERROR: asking for size on UNKNOWN_DATA_TYPE.  0 is what you get.\n";
    return(0);
  default:
    std::cerr << "ERROR: how the heck did we get here?\n";
  }
  return(-1);
}


using namespace std; // be lazy.  FIX:  just include ios?

// With this function, the labels have already been parsed.  need to
// get the image data in a grid
// FIX: what to do about multiband data?
bool
LoadPdsDataFromFile(std::string filename,PdsLabelMap &labels, XcGrid &grd) {
  std::ifstream in(filename.c_str(), ios::in|ios::binary|ios::ate);
  const size_t fileSize=in.tellg();
  //cout << "FileSize = " << fileSize << "\n";

  const int width  = labels["IMAGE-LINE_SAMPLES"].getIntVal();
  const int height = labels["IMAGE-LINES"].getIntVal();
  const dataType type = getDataType(labels);
  const int valueSize = getDataValueSize(type); // Size of one sample

  const int recordSize = labels["RECORD_BYTES"].getIntVal();
  const int imgPointer = labels["^IMAGE"].getIntVal();
  const int startOffset = recordSize * (imgPointer-1);
  const int dataSize = width * height * valueSize;

  //cout << "dataSize="<<dataSize<<" "<<width<<" "<<height<<" "<<valueSize<<"\n";


  if (int(fileSize) <(startOffset+dataSize)) {
    std::cerr << "ERROR: file appears truncated!\n"
	      << "  " << fileSize << " " << startOffset << " " << dataSize<<"\n";
    std::cerr << "  fileSize = " << fileSize<< "\n"
	      << "  startOffset = " << startOffset << "\n"
	      << "  dataSize = " << dataSize<< "\n\n"
	      << "  recordSize = " << recordSize << "\n"
	      << "  imgPointer = " << imgPointer << "\n"
	      << "  width = " << width<< "\n"
	      << "  height = " <<height << "\n"
	      << "  valueSize = " <<valueSize << "\n"
      ;
    return (false);
  }

  in.seekg(startOffset);
  if (in.fail()) {
    std::cerr << "Failed to seek to start of datablock\n";
    return(false);
  }

  char *buf = new char[dataSize+1];
  if (!buf) {
    std::cerr << "Unable to load buffer of "<<dataSize<<"byte.  Out of memory?\n";
    return(false);
  }
  //std::cerr << "Beginning read.\n";
  in.read(buf,dataSize);
  if (in.fail()) {
    std::cerr << "ERROR: unable to read the data block\n";
    delete [] buf;
    return (false);
  }
  //std::cerr << "done with read.\n";

  endian e = getDataEndian(labels);
  //std::cout << "Data Endian: " << e << "\n";
  int x,y;
  for (y=0;y<height;y++) {
#ifndef REGRESSION_TEST    // Things get confused in testing with this
    if (0==y%100) std::cerr << "\r"<< y;
#endif
    for (x=0;x<width;x++) {
	const int i = y * width + x;
	char *ptr = buf + i * valueSize;
	float f;
	f = getFloatVal(ptr, type,  e);
	grd.setZ(x,y,f);
      }
  }
  //std::cerr << "done with loading lines into grd.\n";  
  delete [] buf;

  return(true);
}

/***************************************************************************
 * FUNCTION: AddPdsLabelsToGridLog
 * DESCRIPTION: Put all the labels in the grid log array of strings
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   Only handles straight mxn DEM files.  NO RANGE MAPS!
 ***************************************************************************/
static void AddPdsLabelsToGridLog(XcGrid &grd,PdsLabelMap &labels) {
  PdsLabelMap::iterator i;
  for (i=labels.begin(); i != labels.end();i++) {
    //std::cout << "Next keyword is: " << (*i).first << "\n";
    PdsLabel l = (*i).second;
    grd.addLogEntry(l.getWholeLabelStr());
  }
}

/***************************************************************************
 * FUNCTION: XcGridPdsLoad
 * DESCRIPTION: 
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   Only handles straight mxn DEM files.  NO RANGE MAPS!
 ***************************************************************************/
bool
XcGridPdsLoad (std::string filename,XcGrid &grd) {

  std::vector<std::string> objects;
  PdsLabelMap labels = LoadPdsLabelsFromFile(filename, objects);

  AddPdsLabelsToGridLog(grd,labels);

  //std::cout << "Image line samples:\n";
  const int width  = labels["IMAGE-LINE_SAMPLES"].getIntVal();
  //std::cout << "Image lines:\n";
  const int height = labels["IMAGE-LINES"].getIntVal();

  grd.setRegular(true); // FIX: this will not be the case for range maps
  grd.setSize(width,height);
  grd.setDelXDelY(1.,1.);  	// FIX: what label would these be in?
  grd.setOrigin(0.,0.);		// FIX: what label for origin?

  if (!LoadPdsDataFromFile(filename,labels,grd)) {
    std::cerr << "XcGridPdsLoad: unable to load data\n";
    return(false);
  }

  return (true);
}

//////////////////////////////////////////////////////////////////////
// DRIVER IMPLEMENTATION - XcGridDrv
//////////////////////////////////////////////////////////////////////

std::string XcGridDrvPds::getName() { return("PDS");}
//bool XcGridDrvPds::canReadType(FileType t) {return (false);}
//bool XcGridDrvPds::canWriteType(FileType t) {return(false);}
std::vector<FileType> XcGridDrvPds::getReadTypes() {
  std::vector<FileType> types;
  types.push_back(PDS);
  return (types);
}
std::vector<FileType> XcGridDrvPds::getWriteTypes() {
  std::vector<FileType> types;
  //types.push_back(PDS);
  return (types);
}

bool XcGridDrvPds::ReadGrid(std::string filename, FileType t, XcGrid &grd) {
  assert (PDS==t);
  bool r = XcGridPdsLoad(filename, grd);
  return (r);
}

bool XcGridDrvPds::WriteGrid(std::string filename, FileType t, XcGrid &grd, XcGridOpts &opts) {
  std::cerr << "FIX: No PDS writer implemented yet!\n";
  return (false); // Can't write nuthin.
}

//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////

#ifdef REGRESSION_TEST

// Have to provide out own register command so we don't depend on
// other drivers... Only register ourself.
XcG::VecDriverPtr XcG::XcGridDrvRegister() {
  VecDriverPtr drivers;
  drivers.push_back(new XcGridDrvPds);	// JPL Planetary Data System format
  return(drivers);
}


bool
isequal (const float a, const float b, const float del) {
  if ( a<b+del && a > b-del) return (true);
  return (false);
}

//#include "XcGridPgm.H"

int main (int argc, char *argv[]) {
  bool ok=true;

  if (2!=sizeof(short)) {
    cerr << "  ERROR: short is not 2 bytes.  On a 64bit machine?"
	 << argv[0] << ": FAILED.";
    exit(EXIT_FAILURE);
  }
  if (4!=sizeof(int)) {
    cerr << "  ERROR: int is not 4 bytes.  On a 64bit machine?"
	 << argv[0] << ": FAILED.";
    exit(EXIT_FAILURE);
  }
  if (4!=sizeof(float)) {
    cerr << "  ERROR: float is not 4 bytes.  On a 64bit machine?"
	 << argv[0] << ": FAILED.";
    exit(EXIT_FAILURE);
  }

  if (true) {
    size_t i=0;
    vector<std::string> files;
    files.push_back("TestData/2N126543636ETH0100P1505L0M1.IMG");

    for (i=0;i<files.size();i++) {
      bool loadOK=true;
      XcGrid grd;
      if (!XcGridPdsLoad (files[i],grd)) {
	std::cerr << "ERROR: XcGridPdsLoad failed!\n";
	loadOK = ok = false;
	continue;
      }
      std::cout <<"  Load " << files[i] << (loadOK?": ok.":": FAILED!\n") << std::endl;
    }
  } // Load file 



  ////////////////////////////////////////
  // TEST REGISTRATION
  ////////////////////////////////////////

  {
    VecDriverPtr drivers = XcGridDrvRegister();  // vector of drivers
    bool r = (1==drivers.size());
    if (!r) ok=false;
    std::cout << "  XcGridDrvRegister test: " << (ok?"ok.":": FAILED!") << std::endl;

    // FIX: what is the not found for vector, it's NOT npos??
    if (r) {
      size_t drv;
      drv = XcGridDrvGetReader(drivers,PDS);
      if (string::npos==drv || !drivers[drv]->canReadType(PDS)) {
	std::cout << "  ERROR:  vicar driver failed to respond correctly" << std::endl;
	ok=false;
      }
      //drv = XcGridDrvGetWriter(drivers,PDS);
      //if (string::npos==drv || !drivers[drv]->canWriteType(PDS)) {
      //std::cout << "  ERROR:  vicar driver failed to respond correctly" << std::endl;
      //ok=false;
      //}
    }

    // Test reading.  Btw, we know we have the driver at position 0.
    {
      //XcGridOpts opts;
      XcGrid g;
      std::string filename("TestData/2N126543636ETH0100P1505L0M1.IMG");
      r = drivers[0]->ReadGrid(filename,PDS,g);
      if (!r) ok=false;
      std::cout << "  XcGridDrvRead  test: " << (r?"ok.":"FAILED!") << std::endl;
    }

    // Test write
    if (0) {
      XcGridOpts opts;
      XcGrid g(2,2);
      g.setDelXDelY(1.0,1.0);
      g.setZ(0,0,0);
      g.setZ(1,0,1.1);
      g.setZ(0,1,2.2);
      g.setZ(1,1,3.3);

      std::string filename("test-vicar-write.vic");
      r = drivers[0]->WriteGrid(filename,VICAR,g,opts);
      if (!r) ok=false;
      std::cout << "  XcGridDrvWrite test: " << (r?"ok.":"FAILED!") << std::endl;
    }

    XcGridDrvDelete(drivers);
    r = (0==drivers.size());
    if (!r) ok=false;
    std::cout << "  XcGridDrvDelete test: " << (r?"ok.":"FAILED!") << std::endl;

  } // Registry test


  std::cout << argv[0] << (ok?": ok.":": FAILED!") << std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif // REGRESSION_TEST
