/* $Revision: 1.5 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGridPgm.C
 *
 * DESCRIPTION: Use XcGrid to write/write pgm (pnm) format files.
 *
 * EXPORTS:  
 *
 * NOTES:        Starting off with only the writer
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

//#include <cmath>
#include <iostream>
#include <fstream>
//using namespace std;

#include "XcGrid.H"
#include "XcGridDrv.H"
#include "XcGridPgm.H"

/***************************************************************************
 * FUNCTION: XcGridWritePgmBinary
 * DESCRIPTION: 
 * INPUTS:  
 * 	Max should be either 255 or 65535
 * OUTPUTS: 
 * RETURNS: true on success, false on failure
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 ***************************************************************************/

bool
XcGridWritePgmBinary (const std::string filename, XcGrid &grd, const size_t max) {
  assert (max < 65536);
  assert (255==max);  // FIX: deal with arbitrary size and figure if 1 or 2 byte
  std::ofstream o(filename.c_str(),std::ios::out);
  if (!o) {
    std::cerr << " output file open failed for "<< filename << "\n";
    return (false);
  }

  o << "P5\n"; // Magic Number

  // Write all of the log entries as comments
  {
    o << " # START LOG ENTRIES\n";
    o << " # NumLogEntries = " << grd.getNumLogEntries() << "\n";
    size_t i;
    for (i=0;i<grd.getNumLogEntries();i++) {
      o << " # "
	<< grd.getLogEntry(i)
	<< "\n";
    }
    o << "\n # END LOG ENTRIES\n\n";
  }

  // Image data starts after these thre integers...
  o << grd.getWidth() << " " << grd.getHeight() << "\n"
    << max << "\n";

  float zmin = grd.getZMin();
  float zmax = grd.getZMax();

  size_t x,y;
  for (y=0;y<grd.getHeight();y++) {
    for (x=0;x<grd.getWidth();x++) {
      // FIX: deal with max != 255
      const float z=grd.getZ(x,y);
      float val=round( (z-zmin)/(zmax-zmin)  * 255 );
      unsigned char c = (unsigned char)val;
      o << c;
    }
  }
  o.close();

  return(true);
} // XcGridWritePgmBinary

/***************************************************************************
 * FUNCTION: XcGridWritePgmAscii
 * DESCRIPTION: 
 * INPUTS:  
 * 	Max should be either 255 or 65535
 * OUTPUTS: 
 * RETURNS: true on success, false on failure
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 ***************************************************************************/

bool
XcGridWritePgmAscii (const std::string filename, XcGrid &grd, const size_t max) {
  assert (max < 65536);
  assert (255==max);  // FIX: deal with arbitrary size and figure if 1 or 2 byte
  std::ofstream o(filename.c_str(),std::ios::out);
  if (!o) {
    std::cerr << " output file open failed for "<< filename << "\n";
    return (false);
  }

  o << "P2\n"; // Magic Number for pgm ascii

  // Write all of the log entries as comments
  {
    o << " # START LOG ENTRIES\n";
    o << " # NumLogEntries = " << grd.getNumLogEntries() << "\n";
    size_t i;
    for (i=0;i<grd.getNumLogEntries();i++) {
      o << " # "
	<< grd.getLogEntry(i)
	<< "\n";
    }
    o << "\n # END LOG ENTRIES\n\n";
  }

  // Image data starts after these thre integers...
  o << grd.getWidth() << " " << grd.getHeight() << "\n"
    << max << "\n";

  float zmin = grd.getZMin();
  float zmax = grd.getZMax();

  size_t x,y;
  int entries=0;
  for (y=0;y<grd.getHeight();y++) {
    for (x=0;x<grd.getWidth();x++) {
      // FIX: deal with max != 255
      const float z=grd.getZ(x,y);
      float val=round( (z-zmin)/(zmax-zmin)  * 255 );
      int i = (int)val;
      o << i << " ";
      entries++;
      if (entries > 10) {
	entries = 0;
	o << "\n";
      }
    }	
  }
  o.close();

  return(true);
} // XcGridWritePgmAscii

//////////////////////////////////////////////////////////////////////
// DRIVER IMPLEMENTATION - XcGridDrv
//////////////////////////////////////////////////////////////////////

std::string XcGridDrvPgm::getName() { return("PGM - ascii and binary");}
//bool XcGridDrvPgm::canReadType(FileType t) {return (false);}
//bool XcGridDrvPgm::canWriteType(FileType t) {return(false);}
std::vector<FileType> XcGridDrvPgm::getReadTypes() {
  std::vector<FileType> types;
  return (types);
}
std::vector<FileType> XcGridDrvPgm::getWriteTypes() {
  std::vector<FileType> types;
  types.push_back(PGM_ASCII);
  types.push_back(PGM_BIN);
  return (types);
}

bool XcGridDrvPgm::ReadGrid(std::string filename, FileType t, XcGrid &grd) {
  return (false); // Can't write nuthin.
}
bool XcGridDrvPgm::WriteGrid(std::string filename, FileType t, XcGrid &grd, XcGridOpts &opts) {
  opts.print();
  if (t==PGM_ASCII) {
    bool r;
    r=XcGridWritePgmAscii(filename,grd,opts.maxVal);
    return (r);
  } 
  if (t==PGM_BIN) {
    bool r;
    r=XcGridWritePgmBinary(filename,grd,opts.maxVal);
    return (r);
  } 
  std::cerr << "ERROR XcGridDrvPgm::WriteGrid: bad type." << std::endl;
  return (false);
}


//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////

#ifdef REGRESSION_TEST
int main (int argc, char *argv[]) {
  bool ok=true, r; // r is result for each test.
  {
    XcGrid g(2,2);
    g.setDelXDelY(1.0,1.0);
    g.setZ(0,0,0);
    g.setZ(1,0,100.1);
    g.setZ(0,1,99);
    g.setZ(1,1,255);

    r=XcGridWritePgmBinary("dummy-bin.pgm",g,255);
    std::cout << "  Write " << "dummy-bin.pgm" << (r?": ok.":": FAILED!")<<std::endl;
    if (!r) ok=false;

    r=XcGridWritePgmAscii ("dummy-ascii.pgm",g,255);
    std::cout << "  Write " << "dummy-ascii.pgm" << (r?": ok.":": FAILED!")<<std::endl;
    if (!r) ok=false;
  }

  std::cout << argv[0] << (ok?": ok.":": FAILED!")<<std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif // REGRESSION_TEST
