/* $Revision: 1.9 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGridWritePovRay.C
 *
 * DESCRIPTION: Use XcGrid to write an PovRay mesh2 model
 *
 * EXPORTS:  
 *
 * NOTES:   http://www.povray.org
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

//#include <cmath>
#include <iostream>
#include <fstream>
//using namespace std;

#include "XcGrid.H"
#include "XcGridPovRay.H"

/***************************************************************************
 * FUNCTION: XcGridWritePovRay
 * DESCRIPTION: Create a basic mesh2 model
 * INPUTS:  texture    - true to write texture map
 *	    normals    - true to use XcGrid normals
 *	    colortable - true to use Colortable   FIX: implement
 * OUTPUTS: 
 * RETURNS: true on success, false on failure
 * EXCEPTIONS: 
 * DESIGN:  You will need to create a camera and light file that includes
 *          this model.
 * NOTES:   
 *
 * file:///Applications/POV-Ray 3.5/POV-Ray Mac 3.5 Folder/POV-Ray Help/s_188.html#s06_05_02_04
 *
 ***************************************************************************/


bool
XcGridWritePovRay (const std::string filename, XcGrid &grd,
		   bool texture, bool normals, bool colorTable, 
		   bool writeCamera)
{

  std::ofstream o(filename.c_str(),std::ios::out);
  if (!o.is_open()) {
    std::cerr << " output file open failed for "<< filename << "\n";
    return (false);
  }

  // Write header info...
  o << "// POVRay 3.5 Scene File: " << filename << "\n"
    << "// Produced by Kurt Schwehr's terrain2\n"
    << "#version 3.5;\n\n";

  o << "// Sample render command:\n"
    << "// FIX: strip off .pov from filename"
    << "//   povray +O"<< filename <<".png +W300 +H300 +V -D +X "<< filename <<"\n"
    << "//\n"
    << "\n";


  { // lights camera
    float cx,cy,cz;  	// The point at the center to look at.
    float fx,fy,fz;  	// Point to look from.

    grd.getCenter(cx,cy,cz);

    fx = grd.getXMin() - grd.getWidth()/4;
    fy = grd.getYMax() - grd.getHeight()/2;
    fz = -cz - (grd.getZMax() - grd.getZMin())*2;

    double right[3]; // Right vectors for the camera
    double cam[3] = { fx-cx, fy-cy, fz-cz };
    NormalizeVec (cam);
    double up[3] = { 0.,0.,1.};
    CrossProduct (cam,up,right);


    if (writeCamera) {
      o << "#include \"colors.inc\"\n"
	<< "#include \"textures.inc\"\n"
	<< "camera {location <"<<fx<<" , "<<fy<<" ,"<<fz<<">\n"
	<< "        look_at  <"<<cx<<" , "<<cy<<" , "<<cz<<">\n"
	<< "        up <"<<right[0]<<","<<right[1]<<","<<right[2] <<"> \n"
	<< "        //up <0,0,1> \n"
	<< "        //up <0,1,0> // Default\n"
	<< "       }\n"
	<< "light_source{<1500,2500,-2500> color White}\n"
	<< "\n";
    } else { //no camera
      // Make the user wrap the geometry in a file with camera and lights.
      o << "// SAMPLE OF FILE TO INCLUDE THIS MESH:\n"
	<< "//#include \"colors.inc\"\n"
	<< "//#include \"textures.inc\"\n"
	<< "//camera {location <0.0 , 3.0 ,-3.0>\n"
	<< "//        look_at  <0.0 , 1.0 , 0.0>}\n"
	<< "//light_source{<1500,2500,-2500> color White}\n"
	<< "\n";
    } // writeCamera
  }  // lights camera

  //const int NAN_VAL=-99999;
  //o << "// NAN_VAL = "<<NAN_VAL<<"\n";

  //
  // Setup output stream floating point writing.  Want best precision.
  // 

  o.setf (std::ios_base::fixed,std::ios_base::floatfield);
  const int precision=12;  // FIX: what is really a good number for this?
  o.precision (precision);
  

  o << "mesh2 {\n";

  // WARNING... order matters.  Change the order and PovRay won't read it.

  ////////////////////////////////////////
  // VECTORS
  ////////////////////////////////////////

  //
  // VERTEX_VECTORS
  //

  /*
    vertex_vectors
    {
      number_of_vertices,
      <vertex1>, <vertex2>, ...
    }
  */

  {
    o.precision (precision); // Reset to higher precission
    o << "  vertex_vectors {\n";
    o << "    "<< grd.getNumCompressedIndices() << ", // Num vertices\n";
    size_t i,xoff,yoff,nelemts=grd.getWidth()*grd.getHeight();
    float x,y,z;
    for (i=0;i<nelemts;i++) {
      if (grd.isNaN(i)) continue;
      grd.getXYfromIndex(i,xoff,yoff);
      grd.getXYZ(xoff,yoff,x,y,z);
      // Negative Z to handle the left handed coordinate frame of PovRay
      o << "    <" << x << "," << y << "," << -z << ">,\n";
    }
    o << "  } // vertex_vectors\n\n";
  }

  //
  // NORMAL_VECTORS
  //

  /*
    normal_vectors
    {
      number_of_normals,
      <normal1>, <normal1>, ...
    }
  */
  if (normals) {
    o.precision (5); // FIX: is this enough or too much?
    o << "  normal_vectors {\n";
    o << "    "<< grd.getNumCompressedIndices() << ", // Num vertices\n";
    size_t i,xoff,yoff,nelemts=grd.getWidth()*grd.getHeight();
    float nx,ny,nz;
    for (i=0;i<nelemts;i++) {
      if (grd.isNaN(i)) continue;
      grd.getXYfromIndex(i,xoff,yoff);
      grd.getNormal(xoff,yoff,nx,ny,nz);
      o << "    <" << nx << "," << ny << "," << nz << ">,\n";
    }
    o << "  } // normal_vectors\n\n";
    o.precision (precision); // Reset to higher precission
  }

  //
  // UV_VECTORS - Texture
  //

  /*
    uv_vectors
    {
      number_of_uv_vectors,
      <uv_vect1>, <uv_vect2>, ...
    }
  */


  ////////////////////////////////////////
  // LISTS
  ////////////////////////////////////////

  //
  // TEXTURE_LIST
  //
  /*
    texture_list
    {
      number_of_textures,
      texture { Texture1 },
      texture { Texture2 }, ...
    }
  */

  // Do we need this?

  ////////////////////////////////////////
  // INDICES
  ////////////////////////////////////////

  //
  // FACE_INDICES
  //
  /*
    face_indices
    {
      number_of_faces,
      <index_a, index_b, index_c> [, texture_index [,texture index, texture_index]],
      <index_d, index_e, index_f> [, texture_index [,texture index, texture_index]],
      ...
    }
  */
  {
    size_t i;
    const size_t numTri = grd.getNumTriangles();

    o << "  face_indices\n"
      << "  {\n"
      << "    "<<numTri<<",\n";

    for (i=0; i<numTri; i++) {
      XcTriangle t = grd.getTriangle(i);
      //if (grd.isNaN(i)) continue; // This i has nothing to do with NaN's!
      o << "    <" 
	<< grd.getCompressedIndex(t.vertices[0]) << ","
	<< grd.getCompressedIndex(t.vertices[1]) << ","
	<< grd.getCompressedIndex(t.vertices[2]) << ">,\n";
    }
    o << "  } // face_indices\n";

  } // Coord indexes


  //
  // NORMAL_INDICIES
  //
  /*
    normal_indices
    {
      number_of_faces,
      <index_a, index_b, index_c>,
      <index_d, index_e, index_f>,
      ...
    }
  */

  // I think I can get away without these

  //
  // UV_INDICIES - texture
  //

  /*
    uv_indices
    {
      number_of_faces,
      <index_a, index_b, index_c>,
      <index_d, index_e, index_f>,
      ...
    }
  */

  // I think that I can use the face_indices texture_index instead of this
  // or maybe just leave them totally out and it will default to be correct?

  ////////////////////////////////////////
  // MESH_MODIFIER
  ////////////////////////////////////////

  //
  // inside_vector
  //

  // inside_vector <direction>

  //
  // object_modifiers
  //
  /*
    texture {
      finish { ambient 1 diffuse 0 }
      pigment { image_map { gif "test.gif"  } }
    }
  */

  if (texture) {
    std::string imageType = grd.getTextureFileType();
    o << "  texture {\n"
      << "    finish { ambient 1 diffuse 0.1 }\n"
      << "    pigment { image_map { " << imageType << " \"" << grd.getTextureFile()
      << "\" } }\n"
      << "  } // texture\n";
    //o << "  uv_mapping // Texturing what?";
  } else {
    // no texture
    o << "  pigment {rgb 1} // white\n";
  }

  o << "\n"
    << "} // Mesh2 end for file: " << filename << "\n\n"
    << "// Have a nominal day\n";

  return (true);
} // XcGridWritePovRay


//////////////////////////////////////////////////////////////////////
// DRIVER IMPLEMENTATION - XcGridDrv
//////////////////////////////////////////////////////////////////////

#include "XcGridDrv.H"
using namespace XcG;

std::string XcGridDrvPovRay::getName() { return("POVRAY");}
//bool XcGridDrvPovray::canReadType(FileType t) {return (false);}
//bool XcGridDrvPovray::canWriteType(FileType t) {return(false);}
std::vector<FileType> XcGridDrvPovRay::getReadTypes() {
  std::vector<FileType> types;
  //types.push_back(POVRAY);
  return (types);
}
std::vector<FileType> XcGridDrvPovRay::getWriteTypes() {
  std::vector<FileType> types;
  types.push_back(POVRAY);
  return (types);
}

#if 1
bool XcGridDrvPovRay::ReadGrid(std::string filename, FileType t, XcGrid &grd) {
  return (false); // Can't write nuthin.
}
bool XcGridDrvPovRay::WriteGrid(std::string filename, FileType t, XcGrid &grd, XcGridOpts &opts) {
  return (false); // Can't write nuthin.
}
#endif

//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////

#ifdef REGRESSION_TEST
int main (int argc, char *argv[]) {
  bool ok=true;

  // SIMPLE TEST 1
  if (true){
    XcGrid g(2,2);
    g.setDelXDelY(1.0,1.0);
    g.setZ(0,0,0.01);
    g.setZ(1,0,0);
    g.setZ(0,1,0);
    g.setZ(1,1,0);
    //g.setZToNaN(0,0);
    g.genCompressedIndex();
    g.genTriangles();
    g.genNormals();
    g.genUV();
    if (!XcGridWritePovRay("test1.pov",g,false,true,false,true)) {
      std::cout << "  test 1: FAILED." << std::endl;
      ok=false;
    } else {
      std::cout << "  test 1: ok." << std::endl;
    }
  }



  // SIMPLE TEST 2
  if (true) {
    XcGrid g(2,2);
    g.setDelXDelY(1.0,1.0);
    g.setZ(0,0,0.3);
    g.setZ(1,0,0);
    g.setZ(0,1,0);
    g.setZ(1,1,0);
    g.setZToNaN(1,1);
    g.genCompressedIndex();
    g.genTriangles();
    g.genNormals();
    g.genUV();
    if (!XcGridWritePovRay("test2.pov",g,false,true,false,true)) {
      std::cout << "  test 2: FAILED." << std::endl;
      ok=false;
    } else {
      std::cout << "  test 2: ok." << std::endl;
    }
  }


  // SIMPLE TEST 3
  if (true) {
    XcGrid g(3,4);
    g.setDelXDelY(1.0,-1.0);
    g.setOrigin(0,0);
    g.setZ(0,0,100);
    g.setZ(1,0,110);
    g.setZ(2,0,90);

    g.setZ(0,1,0); g.setZToNaN(0,1);
    g.setZ(1,1,0);
    g.setZ(2,1,0);

    g.setZ(0,2,-50);
    g.setZ(1,2,-80);
    g.setZ(2,2,-90);

    g.setZ(0,3,-150);
    g.setZ(1,3,-180);
    g.setZ(2,3,-190);

    g.scaleZ(0.5,0);
    //g.setZToNaN(1,2);
    //g.setZToNaN(2,2);

    //g.setZToNaN(1,1);

    g.scaleZ(0.01,0);
    g.genCompressedIndex();
    g.genTriangles();
    g.genNormals();
    g.genUV();
    if (!XcGridWritePovRay("test3.pov",g,false,false,false,true)) {
      std::cout << "  XcGridWritePovRay test3.pov: FAILED." << std::endl;
      ok=false;
    } else {
      std::cout << "  XcGridWritePovRay test3.pov: ok." << std::endl;
    }
  }

  // SIMPLE TEST 4
  if (false) {
    XcGrid g(3,3);
    g.setOrigin(-124.593, 41.);
    g.setDelXDelY(0.00724638,-0.00434783);
    g.setZ(0,0,-455.88);
    g.setZ(1,0,-443);
    g.setZ(2,0,-440);

    g.setZ(0,1,-401);
    g.setZ(1,1,-380);
    g.setZ(2,1,-390);

    g.setZ(0,2,-350);
    g.setZ(1,2,-323);
    g.setZ(2,2,-301);

    //g.setZToNaN(1,1);
    g.genCompressedIndex();
    g.genTriangles();
    g.genUV();
    g.scaleZ(1/10000.,0);
    g.genNormals();
    if (!XcGridWritePovRay("test4.pov",g,false,false,false,true)) {
      std::cout << "  XcGridWritePovRay test4.pov: FAILED!" << std::endl;
      ok=false;
    } else {
      std::cout << "  XcGridWritePovRay test4.pov: ok." << std::endl;
    }
  }

  std::cout << argv[0] << (ok?": ok.":": FAILED!") << std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}

#endif // REGRESSION_TEST
