/* $Revision: 1.5 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    xcgridconvert.C
 *
 * DESCRIPTION: Command line program to convert from one type to another
 *	using XcGrid infrastructure.  This is the general interface using
 *
 * EXPORTS:  
 *
 * NOTES:        
 *	Uses XcGridOpts in XcGridDrv.H to pass options.
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

// FIX: list does not work.


/***************************************************************************
 * INCLUDES
 ***************************************************************************/
//#include <cstdio>
//#include <cstdlib>
//#include <cstring>
#include <cassert>

#ifdef IRIX  // If HPUX, MSWindows, IRIX or Solaris, just give up and go home.
#  include "gnugetopt.h"
#else
#  include <getopt.h>
#endif

#include <iostream>


#include "XcGrid.H"
#include "XcGridFile.H" // Contains the type names that we will use
#include "XcGridDrv.H" // Parent for all drivers

/***************************************************************************
 *
 * FUNCTION: printHelp
 * DESCRIPTION: When the user asks for help, give them this
 * INPUTS:  
 * OUTPUTS: Nada
 * RETURNS: Nasa
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   Make sure this stays synced with args!
 *
 ***************************************************************************/

void
printHelp() {
  std::cout << "What!?!? You wanted help!  Silly mortal..." << std::endl;
}

/***************************************************************************
 * MAIN
 ***************************************************************************/

using namespace XcG;
using namespace std;

int main (int argc, char *argv[]) {
  std::string infile;
  std::string outfile;
  std::string outFormatName;
  FileType inFileType = UNKNOWN_FILETYPE;
  FileType outFileType = UNKNOWN_FILETYPE;
  XcGridOpts opts;

  while (1) {
    int c; // get opt return value
    static struct option long_options[] =
      {
	/* These options set a flag. */
	{"verbose", required_argument, 0, 'v'},

	{"in",      required_argument, 0, 'i'},
	{"out",     required_argument, 0, 'o'},
	{"cpt",     required_argument, 0, 'c'},
	{"tex",     required_argument, 0, 't'},
	{"format",  required_argument, 0, 'f'}, // output format
	{"list",    no_argument,       0, 'l'}, // list formats/drivers
	{"normals", no_argument,       0, 'n'}, // force explicit normals
	{"spikes",  no_argument,       0, 's'}, // normal spikes

	//{"xoffset", required_argument, 0, 'x'},
	//{"yoffset", required_argument, 0, 'y'},
	//{"zscale",  required_argument, 0, 'z'},

	{"help",    no_argument,       0, 'h'},
	{0, 0, 0, 0}
      };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    //c=getopt_long (argc, argv, "c:i:o:c:t:x:y:z:h",long_options, &option_index);
    c=getopt_long (argc, argv, "l:v:i:o:t:f:s:n:h",long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    // optarg is a global containing the value.
    std::cout << "FIX: optarg is: '" << optarg <<"'"<< std::endl;
    std::cout.flush();
    switch (c) {
    case 0:
      std::cerr << "What the??? got zero option.  Crap!" << std::endl;
      abort();
      break;

    case 'i': // input file
      infile = std::string(optarg);
      //std::cout << "FIX: Inputfile is:" << infile << std::endl;
      break;

    case 'f': // output format
      outFormatName = std::string(optarg);
      outFileType = XcGridFileType(outFormatName);
      if (UNKNOWN_FILETYPE==outFileType || BAD_FILE==outFileType) {
	std::cerr << "ERROR! Bad file type: " << outFormatName << std::endl;
	// FIX: print list of ok output file types.
	exit(EXIT_FAILURE);
      }
      break;

    case 'h': // help
      printHelp();
      exit (EXIT_FAILURE); // FIX: what should it return???
      break;

    case 'l': // List drivers
      {
	cout << "Here is a list of active drivers:" << endl;
	VecDriverPtr drivers = XcGridDrvRegister();
	XcGridDrvPrint(drivers);
      }
      exit (EXIT_FAILURE); // FIX: what should it return???
      break;

    case 'n': // Force explicite formals in open inventor files.
      cout << "Setting explicit normals" << endl;
      opts.explicitNormals = true;
      break;

    case 'o': // output filename
      outfile = std::string(optarg);
      //std::cout << "FIX: output file is:" << outfile << std::endl;
      break;

    case 's': // debugging spikes in open inventor
      opts.normalSpikes = true;
      break;

    case 't': // texture file
      opts.textureFile = std::string(optarg);
      break;

    case 'v': // verbose mode
      std::cerr << "ERROR verbose mode not implemented yet" << std::endl;
      break;

    case '?': // when we got a bad option.
      std::cerr << "Got a bad option.  Here is some help\n";
      printHelp();
      return (EXIT_FAILURE);
      break;
    default:
      std::cerr << "Option was c = " <<int(c)<< ": '" <<char(c)<< "'" << std::endl;
      std::cerr << "Oh crap... we're going down.  Eject! Eject!" << std::endl;
      std::cerr.flush();
      abort ();
    } // switch
  } // while 1;
  
  inFileType = XcGridGetFileType (infile);

  if (UNKNOWN_FILETYPE==inFileType || BAD_FILE==inFileType) {
    cerr << "ERROR: unable to determine filetype for " << infile << endl;
    return (EXIT_FAILURE);
  }

  VecDriverPtr drivers = XcGridDrvRegister();
  size_t inDrv = XcGridDrvGetReader(drivers,inFileType);
  if (string::npos==inDrv || !drivers[inDrv]->canReadType(inFileType)) {
    cerr << "ERROR: no working driver to read your file" << endl;
    return (EXIT_FAILURE);
  }

  size_t outDrv = XcGridDrvGetWriter(drivers,outFileType);
  if (string::npos==outDrv || !drivers[outDrv]->canWriteType(outFileType)) {
    cerr << "ERROR: no working driver to write your file" << endl;
    return (EXIT_FAILURE);
  }


  XcGrid grd;
  if (!drivers[inDrv]->ReadGrid(infile,inFileType,grd)) {
    cerr << "ERROR: Read failed!" << endl;
    return (EXIT_FAILURE);
  }

  // Sanity check the input file.
  bool ok=true;
  if (grd.getWidth()<=0) {
    cerr << "ERROR: input grid width too small: " << grd.getWidth() << endl;
    ok=false;
  }
  if (grd.getHeight()<=0) {
    cerr << "ERROR: input grid height too small: " << grd.getHeight() << endl;
    ok=false;
  }
  if (!ok) {
    cerr << "Bailing out without writing anything" << endl;
    return (EXIT_FAILURE);
  }

  if (!drivers[outDrv]->WriteGrid(outfile,outFileType,grd,opts)) {
    cerr << "\nERROR: write failed!" << endl;
  }
  
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}

